"""
Provide stubs for verification methods.

The methods listed here will not be automatically transpiled, and their implementation
needs to be provided in form of snippets.

The stubs provide merely the type annotations and hint about the logic.
"""
from typing import TypeVar, Iterable, Set


def is_IRI(text: str) -> bool:
    raise NotImplementedError()


def is_IRDI(text: str) -> bool:
    raise NotImplementedError()


def is_ID_short(text: str) -> bool:
    raise NotImplementedError()


T = TypeVar("T")


def are_unique(iterable: Iterable[T]) -> bool:
    """Return ``True`` if all the elements in the sequence are unique."""
    observed = set()  # type: Set[T]
    count = 0

    for item in iterable:
        observed.add(item)
        count += 1

    return len(observed) == count



"""Define markers for the meta model to mark the functions and data structures."""
from typing import TypeVar, Type, Optional, Tuple

from icontract import require

T = TypeVar("T")


def implementation_specific(thing: Type[T]) -> Type[T]:
    """
    Mark the function or class that can not be defined in the meta-model.

    The specification of the class or function will be disregarded when generating
    the code and serve only as guidance for the developer. Hence the contracts,
    properties and methods need to be manually written.
    """
    return thing


def comment(text: str) -> None:
    """Mark a comment to be included in the generated code."""
    pass


def abstract(thing: Type[T]) -> Type[T]:
    """
    Mark the class as abstract.

    This instructs the code generators for languages which do not support
    multiple inheritance to convert them to interfaces.
    """
    return thing


class reference_in_the_book:
    """Mark the location in the book where the definition resides."""

    @require(lambda section: all(number >= 1 for number in section))
    @require(lambda index: index >= 0)
    def __init__(self, section: Tuple[int, ...], index: int = 0) -> None:
        """
        Initialize with the given values.

        :param section: Section of the book given as tuple (so that it is sortable)
        :param index:
            Index in the section.

            The index helps us distinguish between multiple definitions in a section.
        """
        self.section = section
        self.index = index

    def __call__(self, func: Type[T]) -> Type[T]:
        return func



"""Provide meta-models for Asset Administration Shell information model."""

__version__ = "2021.11.20a2"
__author__ = (
    "Nico Braunisch, Marko Ristin, Robert Lehmann, Marcin Sadurski, Manuel Sauer"
)
__license__ = "License :: OSI Approved :: MIT License"
__status__ = "Alpha"
